unit MagneticForm;

interface

// uTMagneticForm v1.0 - Makes a form "magnetic" to the screen border and
// enables (proportional) resizing and moving on borderless forms
// Delphi implementation (w) 1999 Dennis D. Spreen (dennis@spreendigital.de)
// This unit is freeware. Just drop me a line if this unit is useful for you.

{

TMagneticForm properties
========================

AlwaysMoveable : boolean
		 // true = make a (captionless) form everywhere on the form moveable 
		 // does not work with WndProc-Controls on the form (e.g. TPanels)

AlwaysOnScreen : boolean
		 // true = form can't be moved out of the screen		 

- Margin(TMargin)
   Enabled : boolean (default TRUE)
		 // Enable the "magnetic" screen margin    

   BottomMax : integer (default 10)
   BottomMin : integer (default -5)
   LeftMax   : integer (default 10)
   LeftMin   : integer (default -5)
   RightMax  : integer (default 10)
   RightMin  : integer (default -5)
   TopMax    : integer (default 10)
   TopMin    : integer (default -5)



   A negative value means "outside the visible screen", e.g.:
             .----------------------
"invisible   |
 screen"   . |   .
           | |   |        "visible
           . |   .         screen"
           | |   |        
        x=-5 |   x=10 
       	   | |   |
    	     | 
             x=0 
           \_____/
      
         -5>=x>=10 (form will be moved to x=0 if anywhere between -5 and +10)
     LeftMin    LeftMax         


- Resize (TResize)
   Enabled     : boolean  
                 // true = make a borderless form resizeable

   BorderWidth : integer
                 // "invisible" form border for simulating resize area

   - Ratio (TRatio)
      Enabled  : boolean
		 // true = do a proportional resize only (applies only when
                    user resizes at the edges (TopLeft, TopRight, BottomLeft &
                    BottomRight)

      AspectRatio:single
      Height:integer
      Width:integer
                  // define the proportional value either with an AspectRatio
                  // or by setting initial values for Height and Width

}


uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

  type

  TOnMovingEvent = procedure (Sender: TObject;var rect:TRect) of object;

  TRatio = class (TPersistent)
  private
   fEnabled:boolean;
   fWidth:integer;
   fHeight:integer;
   fAspectRatio:single;
  protected
      procedure SetWidth(value:integer);
      procedure SetHeight(value:integer);
      procedure SetAspectRatio(value:single);
  public
      constructor Create;
      destructor Destroy;override;
  published
   property Enabled:boolean read fEnabled write fEnabled;
   property Width:integer read fWidth write SetWidth;
   property Height:integer read fHeight write SetHeight;
   property AspectRatio:single read fAspectRatio write SetAspectRatio;
  end;

  TResize = class (TPersistent)
  private
      fEnabled:boolean;
      fRatio:TRatio;
      fBorderWidth:integer;
      procedure SetRatio(Value:TRatio);
  public
      constructor Create;
      destructor Destroy;override;
  published
      property Enabled:boolean read fEnabled write fEnabled;
      property Ratio:TRatio read fRatio write SetRatio;
      property BorderWidth:integer read fBorderWidth write fBorderWidth;
  end;


  TMargin = class (TPersistent)
  private
      FLeftMin  :integer;
      FLeftMax  :integer;
      FTopMin   :integer;
      FTopMax   :integer;
      FRightMin :integer;
      FRightMax :integer;
      FBottomMin:integer;
      FBottomMax:integer;
      fEnabled:boolean;
  public
      constructor Create;
      destructor Destroy;override;
  published
      property Enabled:boolean read fEnabled write fEnabled;
      property LeftMin:integer   read FLeftMin   write FLeftMin;
      property LeftMax:integer   read FLeftMax   write FLeftMax;
      property RightMin:integer  read FRightMin  write FRightMin;
      property RightMax:integer  read FRightMax  write FRightMax;
      property TopMin:integer    read FTopMin    write FTopMin;
      property TopMax:integer    read FTopMax    write FTopMax;
      property BottomMin:integer read FBottomMin write FBottomMin;
      property BottomMax:integer read FBottomMax write FBottomMax;
  end;



type
  TMagneticForm = class(TComponent)
  private
    { private declarations }
    fMargin:TMargin;
    fOnMovingEvent:TOnMovingEvent;
    fAlwaysOnScreen:boolean;
    fOldTWndMethod:TWndMethod;
    fAlwaysMoveable:boolean;
    fForm:TForm;
    fResize:TResize;
  protected
    { protected declarations }
    procedure SetMargin(Value:TMargin);
    procedure SetResize(Value:TResize);
    procedure WndProc(var Message: TMessage);
    procedure WMMOVING(var M: TMessage);
    procedure WMNCHitTest(var M: TMessage);
    procedure WMSIZING(var M: TMessage);
  public
    { public declarations }
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
  published
    { published declarations }
    property Margin:TMargin read fMargin write SetMargin;
    property AlwaysOnScreen:boolean read fAlwaysOnScreen write fAlwaysOnScreen;
    property AlwaysMoveable:boolean read fAlwaysMoveable write fAlwaysMoveable;
    property Resize:TResize read fResize write SetResize;
    property OnMoving:TOnMovingEvent read fOnMovingEvent write fOnMovingEvent;
  end;


procedure Register;

implementation


procedure TRatio.SetWidth(value:integer);
begin
 fWidth:=value;
 if Height=0 then fAspectRatio:=0
             else fAspectRatio:=Width/Height;
end;


procedure TRatio.SetHeight(value:integer);
begin
 fHeight:=value;
 if Height=0 then fAspectRatio:=0
             else fAspectRatio:=Width/Height;
end;


procedure TRatio.SetAspectRatio(value:single);
begin
 fAspectRatio:=value;
 fWidth:=100;
 if value=0 then fHeight:=0
            else fHeight:=trunc(100/value);
end;



constructor TRatio.Create;
begin
 inherited Create;
end;

destructor TRatio.Destroy;
begin
 inherited Destroy;
end;


destructor TResize.Destroy;
begin
 fRatio.free;
 inherited Destroy;
end;

constructor TResize.Create;
begin
 inherited Create;
 fRatio:=TRatio.Create;
 fBorderWidth:=5;
end;

procedure TResize.SetRatio(Value:TRatio);
begin
 FRatio.Assign(Value);
end;


constructor TMargin.Create;
begin
 inherited Create;
 LeftMin  :=-5;
 LeftMax  :=10;
 RightMin :=-5;
 RightMax :=10;
 TopMin   :=-5;
 TopMax   :=10;
 BottomMin:=-5;
 BottomMax:=10;
 Enabled:=true;
end;

destructor TMargin.Destroy;
begin
 inherited Destroy;
end;

constructor TMagneticForm.Create(AOwner:TComponent);
begin
 inherited Create(AOwner);
 fMargin:=TMargin.Create;
 fResize:=TResize.Create;
 fForm:=TForm(AOwner);
 fOldTWndMethod:=fForm.WindowProc;
 fForm.WindowProc:=WndProc;
 fResize.Ratio.Width:=fForm.Width;
 fResize.Ratio.Height:=fForm.Height;
 if fForm.BorderStyle=bsNone then AlwaysMoveAble:=true;
end;


destructor TMagneticForm.Destroy;
begin
 fMargin.Free;
 fResize.Free;
 fForm.WindowProc:=fOldTWndMethod;
 // Problems may arise if any component after or before TMagneticform
 // changes the form.WindowProc - the destructor only works if the
 // components are destroyed in the reversed order as constructed
 inherited Destroy;
end;


procedure TMagneticForm.WndProc(var Message: TMessage);
begin
 if (CsDesigning in ComponentState) then fOldTwndMethod(Message) // disable during Delphi IDE
 else
 case Message.Msg of
  WM_MOVING   : WMMOVING(Message);
  WM_NCHITTEST: WMNCHITTEST(Message);
  WM_SIZING   : WMSIZING(Message);
  else fOldTwndMethod(Message);
 end;
end;


procedure TMagneticForm.WMSIZING(var M: TMessage);
var PR : PRect;
begin
 if (not(Resize.Ratio.Enabled)) or (Resize.Ratio.AspectRatio=0) then exit;
 PR := Pointer(m.LParam);
 if M.WParam=WMSZ_BOTTOMRIGHT then PR^.right:=PR^.Left+trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio) else
 if M.WParam=WMSZ_BOTTOMLEFT  then PR^.left :=PR^.Right-trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio) else
 if M.WParam=WMSZ_TOPLEFT     then PR^.left :=PR^.Right-trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio) else
 if M.WParam=WMSZ_TOPRIGHT    then PR^.right:=PR^.Left+trunc((PR^.Bottom-PR^.Top)*Resize.Ratio.AspectRatio);
end;



procedure TMagneticForm.WMMOVING(var M: TMessage);
var PR : PRect;
begin

 PR := Pointer(m.LParam);

 if assigned (OnMoving) then
  begin
   OnMoving(self,pr^);
   fOldTwndMethod(m);
   exit;
  end;

 if ((PR^.left < Margin.LeftMax) and (PR^.Left>Margin.LeftMin) and (Margin.Enabled)) or
    ((AlwaysOnScreen) and (PR^.Left<0))then
  begin
   PR^.left := 0;
   PR^.Right := fForm.width;
  end;

  if ((PR^.top < Margin.TopMax) and (PR^.Top>Margin.TopMin) and (Margin.Enabled)) or
     ((AlwaysOnScreen) and (PR^.Top<0)) then
   begin
    PR^.top := 0;
    PR^.Bottom := fForm.height;
   end;

   if ((PR^.Bottom > screen.Height-Margin.BottomMax) and
      (PR^.Bottom+Margin.BottomMin < screen.Height) and (Margin.Enabled)) or
      ((AlwaysOnScreen) and (PR^.Bottom>screen.height)) then
   begin
    PR^.Bottom := screen.height;
    PR^.top := screen.height - fForm.height;
   end;

   if ((PR^.Right > screen.Width-Margin.RightMax) and
      (PR^.Right + Margin.RightMin< screen.Width) and (Margin.Enabled)) or
      ((AlwaysOnScreen) and (PR^.Right>Screen.Width)) then
   begin
    PR^.Right := screen.width;
    PR^.Left := screen.width - fForm.width;
   end;
 fOldTwndMethod(m);
end;




procedure TMagneticForm.WMNCHitTest(var M: TMessage);
var point:TPoint;
    CPoint:TPoint;
    isLeft,isright,isTop,isbottom:boolean;
begin
  fOldTwndMethod(m); // call old method

  if Resize.Enabled then
   begin
    point.x:=TWMNCHitTest(M).Xpos;
    point.y:=TWMNCHitTest(M).Ypos;

    CPoint:=fForm.ScreenToClient(point);

    isLeft  :=(CPoint.x<Resize.BorderWidth);
    isTop   :=(CPoint.y<Resize.BorderWidth);
    isRight :=(CPoint.x+Resize.BorderWidth>=fForm.ClientWidth);
    isBottom:=(CPoint.y+Resize.BorderWidth>fForm.ClientHeight);

    if IsLeft then if isTop then m.Result:=HTTOPLEFT
                          else
                          if isBottom then m.Result:=HTBOTTOMLEFT
                                      else m.Result:=HTLEFT
             else
    if IsRight then if isTop then m.Result:=HTTOPRIGHT
                           else
                           if isBottom then m.Result:=HTBOTTOMRIGHT
                                       else m.Result:=HTRIGHT
             else
    if IsTop then m.Result:=HTTOP
           else
    if IsBottom then m.Result:=HTBOTTOM;
  end;

  if (m.Result = htClient) and (AlwaysMoveable) then  m.result:=htCaption;
end;



procedure TMagneticForm.SetMargin(Value:TMargin);
begin
 FMargin.Assign(Value);
end;


procedure TMagneticForm.SetResize(Value:TResize);
begin
 FResize.Assign(Value);
end;

procedure Register;
begin
  RegisterComponents('Freeware', [TMagneticForm]);
end;

end.
